#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <stdlib.h>
#include "pngio.h"


int randcol()
{
  int R,G,B;
  R= rand()%256;
  G= rand()%256;
  B= rand()%256;
  return 0xFF000000 | (R<<16) | (G<<8) | B;
}


int scanfil(unsigned char *idata, int iW, int iH, uint32_t **RS);

int main(int argc,char **argv)
{
  char *e; 
  int thr,i;
  char *ofn;
  unsigned char *idata; unsigned int iW, iH;
  uint32_t *parts, nc, *colors;

  srand(time(NULL));

  if (argc!=4)
    exit(fprintf(stderr,"usage: %s input threshold output\n", argv[0]));
  e= read_png_graymap_noalpha(argv[1],&iW,&iH,&idata);
  if (e)
   exit(fprintf(stderr,"can't read %s: %s\n", argv[1], e));
  thr= atoi(argv[2]);
  ofn= argv[3];

  for(i=0;i<iW*iH;i++) if (idata[i]<thr) idata[i]= 0;

  nc= scanfil(idata, iW, iH, &parts);

  colors= malloc(sizeof(uint32_t)*(nc+1));

  colors[0]= 0xFF000000;
  for(i=1;i<nc;i++) colors[i]= randcol();
  for(i=0;i<iW*iH;i++) parts[i]= colors[parts[i]];

  e= write_png_image(ofn, iW, iH, parts);
  if (e)
   exit(fprintf(stderr,"can't write %s: %s\n", ofn, e));
  return 0; 
}

void joinset(uint32_t *M, uint32_t a, uint32_t b)
{
  while(M[a]!=a) a= M[a];
  while(M[b]!=b) b= M[b];

  if (a<b) { int t= a; a= b; b= t; }
  M[a]= b;
#if 0
  // now, a > b

  if (M[a] < b) { M[b]= M[a]; }
   else M[a]= b;
#endif
}

int scanfil(unsigned char *I, int W, int H, uint32_t **RS)
{
  uint32_t *M,*S,*N,R,A,B,C,L;
  M= malloc(sizeof(uint32_t)*(W*H/4+1));
  N= malloc(sizeof(uint32_t)*(W*H/4+1));
  S= calloc(W*H,sizeof(uint32_t));
  int i,x,y,K,lastset;
  M[0]= 0;
  lastset= 0;

#define input(x,y)  I[(x)+(y)*W]
#define output(x,y) S[(x)+(y)*W]

  for(y=0;y<H;y++)
   for(x=0;x<W;x++)
   {
     if (!input(x,y)) continue;
     A= B= C= L= 0;
     if (y!=0)
     {
       if (x!=0 && input(x-1,y-1)) A= output(x-1,y-1);
       if (input(x,y-1))  B= output(x, y-1);
       if (x!=W-1 && input(x+1,y-1)) C= output(x+1,y-1);
     }
     if (x!=0 && input(x-1,y)) L= output(x-1,y);

     if (A==0 && B==0 && C==0)
     {
       if (L==0)
       {
         K= ++lastset;
         output(x,y)= K;
         M[K]= K;
         continue;
       }
       output(x,y)= L;
       continue;
     }

     if (A==0) { if (B==0) R= C; else R= B; } else R= A;

     if (C!=0 && R!=C) { output(x+1,y-1)= R; joinset(M, C, R); }
     if (B!=0 && R!=B) { output(x, y-1)= R; joinset(M, B, R); }
     if (L!=0 && R!=L) { joinset(M, L, R); }
     output(x,y)= R;
   }

   for(i=1;i<=lastset;i++)
     while( M[M[i]]!=M[i] ) M[i]= M[M[i]];

   int la=1;
   N[0]= 0;
   for(i=1;i<=lastset;i++)
     if (M[i]==i) N[i]= la++;

   for(i=0;i<W*H;i++) S[i]= N[M[S[i]]];
   *RS= S;
   free(M); free(N);
   return lastset;
}

